import torch 
torch.set_num_threads(1)
torch.manual_seed(42)

from contraction import symmetrize, Contraction
contract = Contraction.apply

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='')
    parser.add_argument("--D", type=int, default=2, help="D")
    parser.add_argument("--chi", type=int, default=50, help="chi")
    parser.add_argument("--maxiter", type=int, default=50, help="maxiter")
    parser.add_argument("--epsilon", type=float, default=1E-12, help="maxiter")
    parser.add_argument("--nepochs", type=int, default=10, help="nepochs")

    parser.add_argument("--float32", action='store_true', help="use float32")
    parser.add_argument("--use_checkpoint", action='store_true', help="use checkpoint")
    parser.add_argument("--cuda", type=int, default=-1, help="use GPU")
    args = parser.parse_args()
    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64

    d = 2 # fixed
    D = args.D

    B = torch.rand(d, D, D, D, D, dtype=dtype, device=device)
    B = B/B.norm()
    A = torch.nn.Parameter(B)
    
    #dimer covering
    T = torch.zeros(d, d, d, d, d, d, dtype=dtype, device=device)
    T[0, 0, 0, 0, 0, 1] = 1.0 
    T[0, 0, 0, 0, 1, 0] = 1.0 
    T[0, 0, 0, 1, 0, 0] = 1.0 
    T[0, 0, 1, 0, 0, 0] = 1.0 
    T[0, 1, 0, 0, 0, 0] = 1.0 
    T[1, 0, 0, 0, 0, 0] = 1.0 
    T = T.view(d, d**4, d)

    optimizer = torch.optim.LBFGS([A], max_iter=10, tolerance_grad=0, tolerance_change=0, line_search_fn="strong_wolfe") 
    
    def closure():
        optimizer.zero_grad()
        As = symmetrize(A).view(d, D**4)

        T1 = torch.einsum('xa,xby,yc' , (As,T,As)).view(D,D,D,D, d,d,d,d, D,D,D,D).permute(0,4,8, 1,5,9, 2,6,10, 3,7,11).contiguous().view(D**2*d, D**2*d, D**2*d, D**2*d)
        #double layer 
        T2 = (As.t()@As).view(D, D, D, D, D, D, D, D).permute(0,4, 1,5, 2,6, 3,7).contiguous().view(D**2, D**2, D**2, D**2)

        lnT = contract(T1, args.chi, args.maxiter, args.epsilon, args.use_checkpoint)
        lnZ = contract(T2, args.chi, args.maxiter, args.epsilon, args.use_checkpoint)
        loss = (-lnT + lnZ)

        loss.backward()
        return loss
    
    for epoch in range(args.nepochs):
        loss = optimizer.step(closure)
        print ('epoch, loss, gradnorm:', epoch, loss.item(), A.grad.norm().item())
    
    #continue with Newton-CG 
    ####################################################################3
    def fun(x, info):

        A = torch.as_tensor(x, device=device).view(d, D, D, D, D).requires_grad_()
        As = symmetrize(A).view(d, D**4)

        T1 = torch.einsum('xa,xby,yc' , (As,T,As)).view(D,D,D,D, d,d,d,d, D,D,D,D).permute(0,4,8, 1,5,9, 2,6,10, 3,7,11).contiguous().view(D**2*d, D**2*d, D**2*d, D**2*d)
        T2 = (As.t()@As).view(D, D, D, D, D, D, D, D).permute(0,4, 1,5, 2,6, 3,7).contiguous().view(D**2, D**2, D**2, D**2)

        lnT = contract(T1, args.chi, args.maxiter, args.epsilon, args.use_checkpoint)
        lnZ = contract(T2, args.chi, args.maxiter, args.epsilon, args.use_checkpoint)
        loss = (-lnT + lnZ)

        loss.backward(create_graph=True)

        info['feval'] += 1
        info['loss'] = loss
        info['A'] = A

        print (info['feval'], loss.item(), A.grad.norm().item())
        return loss.item(), A.grad.detach().cpu().numpy().ravel()
    
    # see this https://code.itp.ac.cn/wanglei/dpeps/commit/2e47d663bb2c8e155967c4b644edf63d9e9b22e9
    def hessp(x, p, info): 

        A = info['A']
        if not ((x == A.detach().cpu().numpy().ravel()).all()):
            print ('recalculate forward and grad')
            A = torch.as_tensor(x, device=device).view(d, D, D, D, D).requires_grad_()

            As = symmetrize(A).view(d, D**4)

            T1 = torch.einsum('xa,xby,yc' , (As,T,As)).view(D,D,D,D, d,d,d,d, D,D,D,D).permute(0,4,8, 1,5,9, 2,6,10, 3,7,11).contiguous().view(D**2*d, D**2*d, D**2*d, D**2*d)
            T2 = (As.t()@As).view(D, D, D, D, D, D, D, D).permute(0,4, 1,5, 2,6, 3,7).contiguous().view(D**2, D**2, D**2, D**2)
            lnT = contract(T1, args.chi, args.maxiter, args.epsilon, args.use_checkpoint)
            lnZ = contract(T2, args.chi, args.maxiter, args.epsilon, args.use_checkpoint)
            loss = (-lnT + lnZ)

            loss.backward(create_graph=True)
 
        Agrad = A.grad.data.clone() #take A.grad.data out

        #dot it with the given vector
        loss = A.grad.view(-1)@torch.as_tensor(p, device=device)
        A.grad.data.zero_()
        loss.backward(retain_graph=True)

        res = A.grad.detach().cpu().numpy().ravel()
        A.grad.data = Agrad #put it back
        return res

    import scipy.optimize
    x0 = A.detach().cpu().numpy().ravel()
    x = scipy.optimize.minimize(fun, x0, args=({'feval':0},), jac=True, hessp=hessp, method='Newton-CG', options={'xtol':1E-8})
